/*
 * biscom auth client
 * (c) 2009 Klim
 * "THE BEER-WARE LICENSE" 
 * Klim (http://klim.in.ua) wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.                    
 * based on lanauth sources by <visir@telenet.ru>, theMIRON 
 * compile: gcc -O2 -Wall -s -o lanauth lanauth.c
 * run: ./auth -p yourpassword
 */

#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<signal.h>
#include<errno.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<sys/time.h>
#include<sys/socket.h>
#include<unistd.h>
#include<stdio.h>
#include<stdlib.h>
#include<stdarg.h>
#include<string.h>
#include<syslog.h>
#include <netdb.h>
#include "rijndael.h"


static int nodaemon = 0;		/* don't become daemon */
static int ver = 23;			/* protocol version */
static char level = 'a';			/* access level */
static int sock = -1;			/* socket */
static char localip[17] = "0.0.0.0";		/* local ip adress */
static char serverip[17] = "192.168.0.1";	/* gateway ip adress */
static char pass[16] = "";		/* password */
static char passwordIn[16], passwordOut[16]; /* password parts  */

static int recvstr( char *buf, int length);    /* udp receive procedure  */
static void sendstr( char *buf, int length);   /* udp send procedure */
static void opensock();		/* connect to server */
static int auth(void);			/* generate response for v1 protocol */
static void sigusr(int sig);		/* change access level */

static void usage() {
	printf("Usage: auth [-i] [-b localip] [-g gid] [-u uid] [-s serverip] [-l accesslevel] -p password\n");
	exit(0);
}

static void fatal(char *s, ...) {
	va_list ap;
	va_start(ap, s);
	vsyslog(LOG_ERR, s, ap);
	va_end(ap);
	exit(1);
}

int main(int argc, char **argv) {
	char *s;
	int op;
//	unsigned char ch;
//	unsigned char curlevel = 0xff;
	/* process command line arguments */
	while(1) {
		op = getopt(argc, argv, "iv:b:s:p:l:u:g:");
		if (op == -1) break;
		switch(op) {
			case 'i':
				nodaemon = 1;
				break;
			case 'v':
				ver = atoi(optarg);
				break;
			case 'b':
				strncpy(localip, optarg, sizeof(localip)-1);
				localip[sizeof(localip)-1] = 0;
				break;
			case 's':
				strncpy(serverip, optarg, sizeof(serverip)-1);
				serverip[sizeof(serverip)-1] = 0;
				break;
			case 'p':
				strncpy(pass, optarg, sizeof(pass)-1);
				pass[sizeof(pass)-1] = 0;
				/* hide password */
				for (s = optarg; *s; s++) *s = '*';
				break;
			case 'l':
				level = optarg[0];
				break;
			case 'u':
				setuid(atoi(optarg));
				break;
			case 'g':
				setgid(atoi(optarg));
				break;
			default:
				usage();
		}
	}

	if (!pass || !*pass) usage();

	openlog("auth", nodaemon ? LOG_PERROR|LOG_PID : LOG_PID, LOG_DAEMON);
	int i=0;
	for(i=1; i<32; i++) signal(i,sigusr);
//	signal(SIGUSR1, sigusr);
//	signal(SIGUSR2, sigusr);
	if (!nodaemon) {
		signal(SIGPIPE, SIG_IGN);
		signal(SIGHUP, SIG_IGN);
		daemon(0, 0);
	}

  memset(&passwordIn,'Z',16);
  memset(&passwordOut,'0',16);
	strncpy(passwordIn,pass,3);
	strncpy(passwordOut,pass+3,strlen(pass)-3);

	/* main loop */
	opensock();
	while(1) {
		auth();
	}
}

static void sigusr(int sig) {
	switch(sig) {
		case SIGUSR1: level = 'a';
			syslog(LOG_NOTICE,"Switching to Internet mode.....");
			break;
		case SIGUSR2:	level = 'b';
			syslog(LOG_NOTICE,"Switching to Local mode.....");
			break;
		case SIGTERM:
                case SIGINT:
                case SIGQUIT:
                        syslog(LOG_NOTICE,"auth daemon stopped signal %d", sig);
                        closelog();
                        exit(0);
                        break;
                default:
                        syslog(LOG_NOTICE,"got signal %d. ignore...", sig);
                        break;

	}
}

static void sendstr( char *buf, int length)
{
  int clen;
	struct sockaddr_in saddr;
	struct sockaddr *sa;
	sa=&saddr;
	saddr.sin_family=AF_INET;
	saddr.sin_port=htons(7723);
	saddr.sin_addr.s_addr=inet_addr(serverip);
	clen=sizeof(saddr);
	sendto(sock, buf , length ,0 , sa, clen);
}

static int recvstr( char *buf, int length)
{
//	int status;
	struct sockaddr from;
	unsigned int len=sizeof(from);
	return recvfrom(sock, buf, length, 0, &from, &len);
}
static void opensock() {
	struct sockaddr_in sin;
	int proto;
	/* create socket */
	proto=getprotobyname("udp")->p_proto;
	if ((sock = socket(PF_INET, SOCK_DGRAM, proto)) == -1)
	    fatal("socket: %m");

	if (*localip) {
		bzero(&sin, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = inet_addr(localip);
		sin.sin_port = htons(7723);
		if(sin.sin_addr.s_addr == INADDR_NONE)
			fatal("%s: invalid ip address", localip);
		if(bind(sock, (struct sockaddr*)&sin, sizeof(sin)) == -1)
			fatal("bind: %m");
		struct timeval timeout; 
    bzero(&timeout, sizeof(timeout));
		timeout.tv_sec= 20;
		setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&timeout, sizeof(timeout));	
	}
	syslog(LOG_NOTICE, "UDP port 7723 opened\n\tlocal ip:%s\n\tserver ip:%s\n\taccess level:%c\n", localip, serverip, level);
}


static int auth(void) {
  int packetId;
	char buf[200];
	int statusLength;
	char packetIdStr[8];
	struct s_answer {
	  char idSignature[2];
	  char response[16];
	  char packetIdStr[8];
	} answer;
  /* generate random number: */
	srand(time(NULL));
  packetId = rand() % 1000000;  
  memset(buf,0,200);
  sprintf( packetIdStr, "%02d%06d", ver, packetId);
  strcpy(buf,packetIdStr);
  sendstr(&buf,8);
  if (recvstr(&answer,sizeof(answer))<=0)
	{
		syslog(LOG_NOTICE, "Receive timeout");
		sleep(10);	
		return -2;
		
	}  
	if ((answer.idSignature[0]!='i') || (answer.idSignature[1]!='d')){
		syslog(LOG_NOTICE, "Received wrong packet");
		sleep(10);	
	return -1;
  }
	if (strncmp(answer.packetIdStr, packetIdStr,8)!=0){
		syslog(LOG_NOTICE, "Received packet with wrong id %s|%s",&answer.packetIdStr, &packetIdStr);
		sleep(20);	
	  return -1;
	}
   
  RIJNDAEL_context rijndaelContext;
  rijndaelContext.mode = MODE_CBC;
  rijndael_setup(&rijndaelContext, 16, &passwordIn);  
  char key[16], skey[18];
	rijndael_decrypt(&rijndaelContext, &answer.response, &key);
  rijndael_setup(&rijndaelContext, 16, &passwordOut);  
	rijndael_encrypt(&rijndaelContext,  &key, &skey);
	skey[16] = level;
	/*sendstr(&skey,17);*/
	memset(buf, 0, 200);
	sprintf(buf, "%17s%02d%06d", skey, ver, packetId);
	sendstr(buf,25);
	/*memset(buf,0,200);*/
	statusLength = recvstr(&buf,sizeof(buf));
	memset(&rijndaelContext,0,sizeof(rijndaelContext));
	rijndaelContext.mode = MODE_CBC;
	rijndael_setup(&rijndaelContext, 16, &key);  
	int i=0;
	for(i=2;i<statusLength;i+=16)		
		rijndael_decrypt(&rijndaelContext, &buf[i], &buf[i]);
		buf[statusLength]=0;
	syslog(LOG_NOTICE, "Received answer:\n %s",&buf);
	sleep(40);


}

